////////////////////////////////////////////////////////////////////
//!
//! KS-10 Processor
//!
//! \brief
//!      KS-10 Unibus Paging
//!
//! \details
//!     The page table translates Unibus addresses to phyical
//!     addresses.  There are 64 virtual pages which map to 2048
//!     physical pages.
//!
//!       17 16                11 10                 2 1 0
//!       +-+--------------------+--------------------+-+-+
//!       |0| Virtual Page Number|      Word Number   |W|B|
//!       +-+--------------------+--------------------+-+-+
//!                  |6                      |9
//!                  |                       |
//!                 \ /                      |
//!          +-------+-------+               |
//!          |      Page     |               |
//!          |  Translation  |               |
//!          +-------+-------+               |
//!                  |                       |
//!                  |                       |
//!                 \ /11                   \ /9
//!     +------------------------+------------------------+
//!     |  Physical Page Number  |        Word Number     |
//!     +------------------------+------------------------+
//!      16                    26 27                    35
//!
//!
//!     Paging RAM Write
//!
//!              0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!       (LH)  |                                                     |
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!
//!              18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!       (RH)  |FR|ES|FM|VA| 0| 0| 0|             PPN                |
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!
//!
//!     Paging RAM Read
//!
//!              0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!       (LH)  |           | 0|FR|ES|FM|VA|                    | PPN |
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!
//!              18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!       (RH)  |        PPN (cont)        |                          |
//!             +--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+--+
//!
//!     Paging RAM Definitions
//!
//!             FR  : Force Read-Pause-Write (AKA Read Reverse in the
//!                   documents).
//!             ES  : Enable 16-bit Unibus Transfers.  Disable 18-bit
//!                   Unibus transfers.
//!             FM  : Fast Transfer Mode.  In this mode, both odd and
//!                   even words of Unibus data are transferred during
//!                   a single KS10 memory operation (i.e., all 36 bits).
//!             VA  : Page valid.  This bit is set to one when the page
//!                   data is loaded.
//!             PPN : Physical Page Number.
//!
//! \todo
//!
//! \file
//!      ubapag.v
//!
//! \author
//!      Rob Doyle - doyle (at) cox (dot) net
//!
////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2012 Rob Doyle
//
// This source file may be used and distributed without
// restriction provided that this copyright statement is not
// removed from the file and that any derivative work contains
// the original copyright notice and the associated disclaimer.
//
// This source file is free software; you can redistribute it
// and/or modify it under the terms of the GNU Lesser General
// Public License as published by the Free Software Foundation;
// version 2.1 of the License.
//
// This source is distributed in the hope that it will be
// useful, but WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General
// Public License along with this source; if not, download it
// from http://www.gnu.org/licenses/lgpl.txt
//
////////////////////////////////////////////////////////////////////
//
// Comments are formatted for doxygen
//

module UBAPAG(clk, rst, pagWRITE,
              busADDRI, busDATAI, pagDATAO, ubaADDRI, pagADDRO,
              pagVALID, forceRPW, enable16, fastMode);

   input          clk;          // Clock
   input          rst;          // Reset
   input          pagWRITE;     // Write to Paging RAM
   input  [ 0:35] busADDRI;     // KS10 Address to Paging RAM
   input  [ 0:35] busDATAI;     // KS10 Data to Paging RAM
   output [ 0:35] pagDATAO;     // Paging RAM to KS10 Data
   input  [17: 0] ubaADDRI;     // Unpaged Unibus Address (little endian)
   output [16:35] pagADDRO;     // Paging RAM to KS10 Addr
   output         pagVALID;     // Page Valide
   output         forceRPW;     // Force Read/Pause/Write
   output         enable16;     // Enable 16-bit transfers
   output         fastMode;     // Fast Mode

   //
   // Paging RAM
   //
   // Details
   //  The Unibus Paging RAM is 64x15 bits.
   //
   //  The implementation is different than the KS10 implementation.
   //  I have chosen to use a Dual Port RAM which allows the KS10 bus
   //  interface to be implemented on one port (Read/Write) and the
   //  Unibus Paging to be implemented on the second port (Read-only).
   //
   //  This saves a whole ton of multiplexers and uses some free DPRAM.
   //
   // Note
   //  Unibus Paging RAM parity is not implemented.

   //
   // KS10 write to Paging RAM
   //

   wire [0: 5] addr = busADDRI[30:35];
   reg  [0:14] pagRAM[0:63];

   always @(negedge clk)
     begin
        if (pagWRITE)
          pagRAM[addr] <= {busDATAI[18:21], busDATAI[25:35]};
     end

   //
   // KS10 readback of Paging RAM
   //

   assign pagDATAO = {5'b0, pagRAM[addr][0:3], 7'b0, pagRAM[addr][4:14], 9'b0};

   //
   // Unibus Paging
   //
   // Note
   //  See figure in file header for additional information on paging.
   //
   //  Unibus addresses are little-endian.
   //

   wire [0:5] vpn  = ubaADDRI[16:11];           // Virtual Page Number

   assign forceRPW =  pagRAM[vpn][0];
   assign enable16 =  pagRAM[vpn][1];
   assign fastMode =  pagRAM[vpn][2];
   assign pagVALID =  pagRAM[vpn][3];
   assign pagADDRO = {pagRAM[vpn][4:14], ubaADDRI[10:2]};

endmodule
